#!/usr/bin/env python
"""
  Provided here is a template to demonstrate how to run molecular simulations
  to calculate the transition pathways using the module nmftp.

  The computation of the maximum flux transition path requires to calculate
  the conditional expections, mean force of potential and the proto-diffusion
  tensor matrix along path. More specifically, the path is represented by
  J discreted images. 

  As the module nmftp has no insight about the details of molecular system
  except temperature. Therefore, we ask the user to keep track of the path
  in configuration space. We use the ordered index of images to connect the
  path in collective variable space and configuration space.

  The first step in general is to prepare proper (equilibrated) system at each
  image such that the restraints are satisfied.
  The function "preprocess" serves for doing this.

  The second step is the main step, in which the user should provide a function,
  called mean_force to serve as an interface with the module nmftp. 
  The molecular simulators are often designed to communicate through I/O, and
  it is often suggested to run several short trajectories.
  The function mean_force starts with running a period of equilibrating the
  system, and is followed by sampling through several short runs.

  The last step is to calculate statistical quantities, including the flow rate,
  isocommittor, potential of mean force, transition path in collective variable
  space and in configuration space (having a subdirectory named CRD as an
  example in this program)

  INPUT contains all the script files needed for simulator
  OUTPUT is to store intermediate quantities
  RESTART is to save the restart files for short runs  
"""

import subprocess as sp
import numpy as np
from math import pi
import time
import os

try:
    from mpi4py import MPI
except:
    try:
        import FMPI as MPI  # FAKE mpi4py
    except:
        print "MPI4PY NOT FOUND"

from nmftp import refine_path

def main():
    start = time.time()

    # make auxiliary directories CRD and OUTPUT 
    myrank = MPI.COMM_WORLD.Get_rank()
    if myrank == 0:
        # make directorys CRD, OUPUT, RESTART
        if not os.path.isdir('CRD'):
            os.mkdir('CRD')
        if not os.path.isdir('OUTPUT'):
            os.mkdir('OUTPUT')
        if not os.path.isdir('RESTART'):
            os.mkdir('RESTART')
    MPI.COMM_WORLD.Barrier()

    # initialize molecular system object
    msys = MSys()
    
    # specify system temperature
    msys.temp = 300

    # define the control parameters: stopping criteria, accuracy, performance 
    misc = {'TOL':0.01, 'noiter':2, 'kcst': 2000.0, 'tau':0.1}  

    # initialize a path in collective variable space
    initpath = []
    if myrank == 0:
        initpath = np.loadtxt('initpath.dat')
    initpath = MPI.COMM_WORLD.bcast(initpath, root = 0)
    
    # initialize a path in Cartesian space: create a directory CRD in which,
    # for each image of the path in collective variable space, there is a
    # one-to-one crystal structure lined by index number along the path
    # in Cartesian space satisfying the restraints from the images along
    # the path in collective variable pace.
    msys.preprocess(initpath, misc['kcst'])

    # From now on, units are in radius rather than in degrees. In general,
    # this is not necessary and is problem-dependent.
    initpath = (pi/180)*initpath

    # calling refine_path from nmftp module.
    # The current method returns the transition path and performance estimates. 
    # potential of mean force, committor, and flow rate will be included soon.
    [tranpath] = refine_path(msys, initpath, misc)

    duration = time.time() - start
    
    # print, save and visualize data
    if myrank == 0:
        tranpath = (180/pi)*tranpath
        initpath = (180/pi)*initpath
        print 'It takes %d iterations %f seconds to converge with error %f.'\
              %(misc['noiter'], duration, misc['TOL'])
        np.savetxt('tranpath.dat', tranpath)
        try:
            import pylab as p
            p.figure()
            p.plot(initpath[:,0], initpath[:,1], 'r-o',
                   tranpath[:,0], tranpath[:,1], 'k-s')
            p.show()
        except:
            print 'Pylab is not installed. The transition path is saved.'
    MPI.Finalize()

# Define a molecular system.
# This should include temperature, script files (minimization, equilibration,
# sampling), definition of collective variables and relavent masses.
# For our convenience, it also includes the spring constant of harmonic restrain.
# The key method is mean_force, which is the interface with the module nmftp.
class MSys:

    def __init__(self):
        self.temp = 300.0  
        self.min = 'INPUT_CHARMM/adp_min.inp'
        self.equil = 'INPUT_CHARMM/adp_equil.inp'
        self.sampling = 'INPUT_CHARMM/adp_sampling.inp'
        self.colvar = ['phi', 'psi']
        self.mass = [12.0110,14.0070,12.0110,12.0110,14.0070]
    # The key method 'mean_force' has to be defined, providing mean force F and
    # proto-diffusion tensor matrix D
    def mean_force(self, zeta, j, kcst):
        dim_cv = len(self.colvar)
        F = np.linspace(0.0, 0.0, dim_cv)
        D = np.zeros((dim_cv, dim_cv),'float')
        DP = np.zeros((dim_cv, dim_cv, dim_cv),'float')

        # convert radius to degree
        zeta = fold(np.array(zeta)*180/pi)
        beta = 1.67*300/self.temp
        charmm_k = kcst/2.0

        #####################################################################
        # Run molecular simulator to generate trajectories under the harmonic
        # restraints with center (phi0, psi0) and force kcst
        # The input coordinates and velocities are restart files with index
        # number j from the directory RESTART. 
        #####################################################################

        # Equilibrate molecular system for 5 ps
        # The input configuration (positions) is read from directory CRD
        # and output configurations (positions and velocities) in a format
        # of restart files are written into directory RESTART.
        # In addition, the configuration (positions) in CRD will be overwritten
        # by the configuration at the last step of equilibration.
        estep = 500
        cwd = os.path.abspath('') + '/'
        inp = open(cwd + self.equil,'r')
        out = open(cwd + "OUTPUT/equil_%d.out"%j,'w')
        p1 = sp.Popen(["charmm", "strid=%d"%j,
                       self.colvar[0] + '=%f'%zeta[0],
                       self.colvar[1] + '=%f'%zeta[1],
                       "temp=%d"%self.temp, "k=%f"%charmm_k, "es=%d"%estep],
                      stdin=inp, stdout=out, stderr=out)
        p1.wait()
        out.close()
        inp.close()
        
        # Run sampling for 5*N pico seconds
        N = 2
        for i in range(N):
            sp.call(["rm -f " + cwd + "OUTPUT/*_%d.*"%j,
                     "rm -f " + cwd + "RESTART/file_%d.res"%j,
                     "mv "+cwd+"RESTART/temp_%d.res RESTART/file_%d.res"%(j,j)],
                    shell = True)
            # The input configurations (positions and velocities) are read from
            # directory RESTART.
            # The trajectories of selected atoms are written into files for
            # communication with python scripts. (This will be done using the
            # named pipe feature)
            # The restart files in directory RESTART will be overwritten after
            # each short run.
            pstep = 500
            inp = open(cwd + self.sampling,'r')
            out = open(cwd + "OUTPUT/sampling_%d.out"%j,'w')
            p1 = sp.Popen(["charmm", "strid=%d"%j,
                       self.colvar[0] + '=%f'%zeta[0],
                       self.colvar[1] + '=%f'%zeta[1],
                           "temp=%d"%self.temp,"k=%f"%charmm_k,"rs=%d"%pstep],
                          stdin=inp, stdout=out, stderr=out)
            p1.wait()
            out.close()
            inp.close()
            
            # load the trajectories of atoms
            CLP = np.loadtxt(cwd + 'OUTPUT/clp_%d.dat'%j)
            NL = np.loadtxt(cwd + 'OUTPUT/nl_%d.dat'%j)
            CA = np.loadtxt(cwd + 'OUTPUT/ca_%d.dat'%j)
            CRP = np.loadtxt(cwd + 'OUTPUT/crp_%d.dat'%j)
            NR = np.loadtxt(cwd + 'OUTPUT/nr_%d.dat'%j)

            ##################################################
            # Analyze the trajectory
            ##################################################  
            # calculate dihedrals phi and psi
            phi = torsion(CLP, NL, CA, CRP)
            psi = torsion(NL, CA, CRP, NR)
            # calcuate Jacobians of \partial phipartial x, and
            # \partial psi/partial x
            dphi = derivativetorsion(CLP, NL, CA, CRP)
            dpsi = derivativetorsion(NL, CA, CRP, NR)
            # calculate the difference
            phi = phi - zeta[0]
            psi = psi - zeta[1]
            # the difference should be less than 180
            phi = phi - np.fix(phi/180)*360
            psi = psi - np.fix(psi/180)*360
            # convert to radius
            phi = phi*pi/180
            psi = psi*pi/180
            # masses of five atoms for alanine dipeptide
            m = [12.0110,14.0070,12.0110,12.0110,14.0070]
            # calculate xi_x M^{-1} xi_x^T
            dphi2 = dphi*dphi
            dpsi2 = dpsi*dpsi
            dphipsi = dphi[:,3:12]*dpsi[:,0:9]

            temp = (dphi2[:,0]+dphi2[:,1]+dphi2[:,2])/m[0] + \
                   (dphi2[:,3]+dphi2[:,4]+dphi2[:,5])/m[1] + \
                   (dphi2[:,6]+dphi2[:,7]+dphi2[:,8])/m[2] + \
                   (dphi2[:,9]+dphi2[:,10]+dphi2[:,11])/m[3]
            D[0,0] += np.mean(temp)
            DP[0,0,0] += np.mean(temp*phi)
            DP[1,0,0] += np.mean(temp*psi)

            temp = (dphipsi[:,0]+dphipsi[:,1]+dphipsi[:,2])/m[1] + \
                   (dphipsi[:,3]+dphipsi[:,4]+dphipsi[:,5])/m[2] + \
                   (dphipsi[:,6]+dphipsi[:,7]+dphipsi[:,8])/m[3]  
            D[1,0] += np.mean(temp)
            DP[0,1,0] += np.mean(temp*phi)
            DP[1,1,0] += np.mean(temp*psi)

            temp = (dpsi2[:,0]+dpsi2[:,1]+dpsi2[:,2])/m[1] + \
                   (dpsi2[:,3]+dpsi2[:,4]+dpsi2[:,5])/m[2] + \
                   (dpsi2[:,6]+dpsi2[:,7]+dpsi2[:,8])/m[3] + \
                   (dpsi2[:,9]+dpsi2[:,10]+dpsi2[:,11])/m[4]  
            D[1,1] += np.mean(temp)
            DP[0,1,1] += np.mean(temp*phi)
            DP[1,1,1] += np.mean(temp*psi)

            F[0] += np.mean(phi)
            F[1] += np.mean(psi)

        D[0,1] = D[1,0]
        DP[0,0,1] = DP[0,1,0]
        DP[1,0,1] = DP[1,1,0]

        F = -(kcst/N)*F
        D = (1/(2.0*beta*N))*D
        for i in range(dim_cv):
            DP[i] = beta*F[i]*D + (kcst/(2*N))*DP[i]

        return [F, D, DP]

    # Initialize an initial path in Cartesian space by
    # minimizing, heating, and equilibrating.
    def preprocess(self, path, kcst, mstep = 1000, estep = 1000):

        cwd = os.path.abspath('') + '/'
        no_img = len(path)
        beta = 1.67*300/self.temp
        charmm_k = kcst/2.0

        ##################################################
        # Run molecular simulator to minimize
        ##################################################

        # minimize one of two end images of the discretized path 
        # In this example, we are provided with a crystal structure of alanine
        # dipeptide and an initial path in phi-psi space.
        # The first step is to minimize the molecular system under the harmonic
        # restrains of phi-psi with spring constant k.
        # The input configuration (positions) is the crystal structure from
        # directory INPUT, and the minimized configuration (positions) is
        # written to directory CRD.
        coor = cwd + "INPUT_CHARMM/aladip.crd"
        inp = open(cwd + self.min,'r')
        out = open(cwd + "OUTPUT/min_%d.out"%0,'w')
        p1 = sp.Popen(["charmm", "coord=%s"%coor, "strid=%d"%0,
                       self.colvar[0] + '=%f'%path[0][0],
                       self.colvar[1] + '=%f'%path[0][1],
                       "k=%f"%charmm_k, "ms=%d"%mstep],
                      stdin=inp, stdout=out, stderr=out)
        p1.wait()
        out.close()
        inp.close()
        
        # minimize the others sequentially
        for j in range(1, no_img):
            coor = cwd + "CRD/adp_%d.crd"%(j-1)
            inp = open(cwd + self.min,'r')
            out = open(cwd + "OUTPUT/min_%d.out"%j,'w')
            p1 = sp.Popen(["charmm", "coord=%s"%coor, "strid=%d"%j,
                           self.colvar[0] + '=%f'%path[j][0],
                           self.colvar[1] + '=%f'%path[j][0],
                           "k=%f"%charmm_k, "ms=%d"%mstep],
                          stdin=inp, stdout=out, stderr=out)
            p1.wait()
            out.close()
            inp.close()
            
        ##################################################
        # Run molecular simulator to heating/equilibrating
        ##################################################

        noproc = MPI.COMM_WORLD.Get_size()
        myrank = MPI.COMM_WORLD.Get_rank()
        noloop = (no_img-1)/noproc +1
        for j in range(myrank, no_img, noproc):
        # The minimized molecular system need to be heated up and equilibrated.
        # The input configuration (positions) is read from directory CRD
        # and output configurations (positions and velocities) in a format
        # of restart files are written into directory RESTART.
        # In addition, the configuration (positions) in CRD will be overwritten
        # by the configuration at the last step of equilibration.
            coord = cwd + "CRD/adp_%d.crd"%(j)
            inp = open(cwd + self.equil,'r')
            out = open(cwd + "OUTPUT/equil_%d.out"%j,'w')
            p1 = sp.Popen(["charmm", "coord=%s"%coor, "strid=%d"%j,
                           self.colvar[0] + '=%f'%path[j][0],
                           self.colvar[1] + '=%f'%path[j][0],
                           "temp=%f"%self.temp, "k=%f"%charmm_k, "es=%d"%estep],
                          stdin=inp, stdout=out, stderr=out)
            p1.wait()
            out.close()
            inp.close()

# convert to the range [-180,180] for an angle
def fold(coors):
    """Fold coors"""
    coors = np.mod(coors, 360)
    coors -= np.floor((coors+180)/360)*360

    return coors

#
# calculate the torsion angle defined by r1, r2, r3, and r4
# np.sum(x*y,axis=1) is replace by np.dot(x,y)/np.inner(x,y)
# for calculating only a torsion angle
#
def torsion(r1, r2, r3, r4):
    d1 = r2 - r1
    d2 = r3 - r2
    d3 = r4 - r3 
    cross = np.cross(d2, d3)

    if np.size(r1) == 3: # one torsion angle
        d12 = np.inner(d1,d2)
        d22 = np.inner(d2,d2)
        d13 = np.inner(d1,d3)
        d23 = np.inner(d2,d3)
        sphi = np.sqrt(d22)*np.inner(d1,cross)
    else: # multiple torsion angles
        d12 = np.sum(d1*d2,axis=1)
        d22 = np.sum(d2*d2,axis=1)
        d13 = np.sum(d1*d3,axis=1)
        d23 = np.sum(d2*d3,axis=1)
        sphi = np.sqrt(d22)*np.sum(d1*cross,axis=1)
    cphi = d12*d23 - d22*d13
    phi = 45.0*np.arctan2(sphi, cphi)/np.arctan(1.0)

    return phi # return in unit degree

#
# calculate the "derivative of a torsion angle", i.e.,
# the Jacobian matrix defined by $\partial phi /partial r
# where phi is denoted as the torsion angle.
# np.sum(x*y,axis=1) is replace by np.dot(x,y)/np.inner(x,y)
# for calculating only a torsion angle
#
def derivativetorsion(r1, r2, r3, r4):
    jacobian = np.zeros((np.size(r1)/3,12),'float')

    F = r1 - r2
    G = r2 - r3
    H = r4 - r3
    A = np.cross(F, G)
    B = np.cross(H, G)

    if np.size(r1) == 3: # one torsion angle
        AA = np.inner(A, A)
        BB = np.inner(B, B)
        FG = np.inner(F, G)
        HG = np.inner(H, G)
        Gnorm = np.sqrt(np.inner(G, G))
        jacobian[0,0:3] = -(Gnorm/AA)*A
        jacobian[0,3:6] = (Gnorm/AA)*A + (FG/(AA*Gnorm))*A - (HG/(BB*Gnorm))*B
        jacobian[0,6:9] = (HG/(BB*Gnorm))*B - (FG/(AA*Gnorm))*A - (Gnorm/BB)*B
        jacobian[0,9:12] = (Gnorm/BB)*B
    else: # multiple torsion angles
        AA = np.sum(A*A, axis=1)
        BB = np.sum(B*B, axis=1)
        FG = np.sum(F*G, axis=1)
        HG = np.sum(H*G, axis=1)
        Gnorm = np.sqrt(np.sum(G*G, axis=1))
        for k in range(3):
            jacobian[:,k] = -(Gnorm/AA)*A[:,k]
            jacobian[:,k+3] = (Gnorm/AA)*A[:,k] + (FG/(AA*Gnorm))*A[:,k] \
                              - (HG/(BB*Gnorm))*B[:,k]
            jacobian[:,k+6] = (HG/(BB*Gnorm))*B[:,k] - (FG/(AA*Gnorm))*A[:,k] \
                              - (Gnorm/BB)*B[:,k]
            jacobian[:,k+9] = (Gnorm/BB)*B[:,k]

    return jacobian # [pri, prj, prk, prl]

if __name__ == '__main__':
    main()
